_G["LootSquid"] = {
    components = {},
    session = {
        id = nil, -- unique key for every "offer to raid" click
        loot = {},
        currentLootSlot = 0,
        playersNeeding = {},
        playersGreeding = {}
    },
    debug = {
        alwaysParseLoot = false,
        alwaysParseChat = false,
        logging = false,
        disableWhispers = false,
        addonComm = true,
        addonCommDebugee = { "Viggò-TarrenMill", "Viggo-TarrenMill", "Vïggo-TarrenMill", "Vìx-TarrenMill", "Sixpack-Bloodfeather" }
    },
    settings = {
        lootThreshold = 0,
        layout = {
            windowFrame = {
                width = 0.65, -- 1/1000
                height = 0.65 -- 1/1000
            },
            requestFrame = {
                playerInfo = 0.06, -- 1/1000
                currentItem = 0.16, -- 1/1000
                improvement = 0.26, -- 1/1000
                data1 = 0.36, -- 1/1000
                data2 = 0.44, -- 1/1000
                data3 = 0.52, -- 1/1000
                data4 = 0.60, -- 1/1000
                data5 = 0.68, -- 1/1000
                votes = 0.8,
                actions = 0.94 -- 1/1000
            }
        }
    },
    voting = {
        requested = false,
        sessionId = nil,
        votes = {--["Viggò"] = { votedBy = {}, comment = "gg" },
        }
    },
    frames = { mainFrame = nil },
    eventHandlers = {},
    data = { maxStanding = 0 },
    lastChatRequestEventTime = 0,
    lastGetItemInfoEventTime = 0,
    lastRequestFrameUpdatedTime = 0,
    versionMajor = "0",
    versionMinor = "4",
    versionRevision = "0",
    versionState = "alpha",
    versionNewNotificationTriggered = false
}

StaticPopupDialogs["LOOTSQUID_AWARD_LOOT"] = {
    text = "Transfer loot?",
    button1 = "Yes",
    button2 = "No",
    OnAccept = function(popup)

        -- find candidate for current loot slot
        local shortPlayerName = Ambiguate(popup.playerName, "none")
        for k in pairs(_G.LootSquid.session.loot) do
            if (_G.LootSquid.session.loot[k].candidates
                    and _G.LootSquid.session.loot[k].candidates[shortPlayerName]) then

                local ci = _G.LootSquid.session.loot[k].candidates[shortPlayerName].candidateIndex

                if ci > 0 then
                    LootSquid:TransferLoot(_G.LootSquid.session.currentLootSlot, ci)
                    LootSquid:ClearNeedAndGreedRequests()
                end
            end
        end

        --
    end,
    timeout = 30,
    whileDead = false,
    hideOnEscape = true,
    preferredIndex = 3, -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
}

StaticPopupDialogs["LOOTSQUID_WARNING_DATA_OLD"] = {
    text = "|cFFFF0000LootSquid data is more than half a day old. Please update external data.",
    button1 = "OK!",
    timeout = 600,
    whileDead = true,
    hideOnEscape = true,
    preferredIndex = 3, -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
}

function LootSquid:OnLoad(mainFrame)
    mainFrame:ClearAllPoints()
    mainFrame:SetPoint("LEFT", 0)
    mainFrame:SetPoint("TOP", 0)
    self.frames.mainFrame = mainFrame
    self:Initialize()
end

function LootSquid:Initialize()

    local components = {
        ["LootSquid Data"] = LootSquidData,
        ["LootSquid Utilities"] = LootSquid.components.utils,
        ["LootSquid Events"] = LootSquid.components.events,
        ["LootSquid Information Logging"] = LootSquid.LogInfo,
        ["LootSquid Error Logging"] = LootSquid.LogError
    }

    for k, v in pairs(components) do
        if not v then
            DEFAULT_CHAT_FRAME:AddMessage(format("|cFFFFAA00%s:|cFFFF0000 %s", "LootSquid [Error]",
                "Missing Component: "
                        .. tostring(k)
                        .. ". Please make sure addon is installed correctly."))
        end
    end

    self:RegisterEvents()


    if _G["LootSquidData"] then
        if (_G["LootSquidData"].data
                and _G["LootSquidData"].dataHash
                and _G["LootSquidData"].dataHashUnixTimeStamp
        and not _G["LootSquidData"].ignoreData
        ) then
            local success, message = pcall(function()
                LootSquid:LogInfo("Successfully loaded data. Timestamp (local): "
                        .. LootSquid:TimeToDateStr(_G["LootSquidData"].dataHashUnixTimeStamp)
                        .. ", Hash: " .._G["LootSquidData"].dataHash)
            end)

            if not success then
                LootSquid:LogError(
                    "Failed to load data. Make sure dataHashUnixTimeStamp is a number corresponding to a unix timestamp. "
                    .. "Error: " .. tostring(message))
            end
        else
            LootSquid:LogError("No or invalid data available in LootSquidData (maybe an old data file?).")
            _G["LootSquidData"] = nil
        end
    else
        LootSquid:LogError("No external data available. Probably missing LootSquidData.lua file.")
    end


    -- create icon
    local iconStr = "Interface\\Icons\\Spell_Paladin_HammerOfWrath"

    local iconTextureAndLabelWrapper = LootSquid:CreateFrameSolid(UIParent, 80, 80,
        { r = 0.1, g = 0.1, b = 0.1, a = 0.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 0.0 },
        3, true)

    iconTextureAndLabelWrapper:SetSize(80, 80)
    iconTextureAndLabelWrapper:SetRelativePoint("CENTER", "CENTER", 0, 0.2)
    iconTextureAndLabelWrapper:SetFrameStrata("FULLSCREEN")

    iconTextureAndLabelWrapper:HookScript("OnMouseUp",
        function(selfFrame, button)
            if button == "RightButton" and not selfFrame.isMoving then
                LootSquid:OpenLootGui()
            end
        end)

    self.frames.iconTextureAndLabelWrapper = iconTextureAndLabelWrapper

    local iconTextureItem = iconTextureAndLabelWrapper:CreateTexture()
    iconTextureItem:SetSize(40, 40)
    iconTextureItem:SetPoint("CENTER", 0, 10)
    iconTextureItem:SetTexture(iconStr)
    self.frames.iconTextureItem = iconTextureItem

    local iconLabelItem = LootSquid:CreateLabelStandard(iconTextureAndLabelWrapper,
        "Loot Squid",
        nil, "GameFontWhiteTiny")
    iconLabelItem:SetPoint("CENTER", 0, -15)
    iconLabelItem:SetWidth(80)
    self.frames.iconLabelItem = iconLabelItem


    iconTextureAndLabelWrapper:Show()

end

function LootSquid:TimerFeedback()
    print("5 seconds passed")
end

function LootSquid:LootIconClick(lootIndex)
    --LootSquid:LogInfo("Selecting lootIndex: " .. tostring(lootIndex))
end

function LootSquid:LootIconEnter(itemLink)
    GameTooltip:SetOwner(self.frames.window, "ANCHOR_CURSOR")
    GameTooltip:SetHyperlink(itemLink)
    GameTooltip:Show()
end

function LootSquid:LootIconLeave()
    GameTooltip:Hide()
end

function LootSquid:TransferLoot(lootSlot, candidateIndex)
    GiveMasterLoot(lootSlot, candidateIndex)
end

function LootSquid:InitGui()

    -- create new window

    -- settings
    local headerHeight = 30
    local footerHeight = 20
    local padding = 3
    local paddingRel = 0.005

    --local horizontalGuiScale = LootSquid.settings.layout.windowFrame.width / 1000.0
    --local verticalGuiScale = LootSquid.settings.layout.windowFrame.height / 1000.0

    local f = self:CreateFrameSolid(UIParent,
        LootSquid.settings.layout.windowFrame.width,
        LootSquid.settings.layout.windowFrame.height,
        { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        padding,
        true)

    f:SetPoint("CENTER", 0, 0)
    f:SetFrameStrata("FULLSCREEN")

    f:SetScript("OnKeyDown",
        function(self, key)
            if key == "ESCAPE" then
                f:SetPropagateKeyboardInput(false)
            else
                f:SetPropagateKeyboardInput(true)
            end
        end)

    f:SetScript("OnKeyUp",
        function(self, key)
            if key == "ESCAPE" then
                f:Hide()
            end
        end)

    local header = self:CreateFrameSolid(
        f,
        nil,
        nil,
        { r = 0.12, g = 0.12, b = 0.12, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    header:ClearAllPoints()
    header:SetPoint("TOPLEFT", f, "TOPLEFT", padding, -padding)
    header:SetPoint("BOTTOMRIGHT", f, "TOPRIGHT", -padding, -headerHeight-padding)
    --header:SetRelativePoint("BOTTOMRIGHT", "BOTTOMRIGHT", -paddingRel, 0.9 - paddingRel)

    self.frames.headerFrame = header

    local b = self:CreateButtonStandard(header, 25, 25, "x")
    b:SetPoint("RIGHT", -padding, 0)
    b:SetPoint("TOP", 0, -padding)
    b:SetScript("OnClick", function(selfButton, button) f:Hide() end)

    local title = self:CreateLabelStandard(header,
        string.format("Loot Squid - v%d.%d.%d %s", self.versionMajor, self.versionMinor, self.versionRevision, self.versionState))
    title:SetPoint("LEFT", 8, 0)

    local footer = self:CreateFrameSolid(f,
        nil,
        nul,
        { r = 0.12, g = 0.12, b = 0.12, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    footer:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT", padding, padding)
    footer:SetPoint("TOPRIGHT", f, "BOTTOMRIGHT", -padding, padding+footerHeight)


    self.frames.footerFrame = footer

    LootSquid.frames.footerLabel = self:CreateLabelStandard(footer,
        "footerLabel",
        nil, "GameFontWhiteTiny")

    LootSquid.frames.footerLabel:SetRelativePoint("CENTER", "CENTER", 0, 0)

    local bodyFrame = self:CreateFrameSolid(f,
        (f:GetWidth() - padding * 2),
        (f:GetHeight() - padding * 4 - headerHeight - footerHeight),
        { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    bodyFrame:SetPoint("TOP", 0, -padding - headerHeight - padding)
    bodyFrame:SetPoint("LEFT", padding, 0)

    self.frames.bodyFrame = bodyFrame

    local lootFrame = self:CreateFrameSolid(bodyFrame,
        (bodyFrame:GetWidth() - padding) / 2 - 16 - padding,
        (bodyFrame:GetHeight() - padding * 2) / 3,
        { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    lootFrame:SetPoint("TOP", 0, 0)
    lootFrame:SetPoint("LEFT", 0, 0)

    self.frames.lootFrame = lootFrame

    local lootFrameTitle = self:CreateLabelStandard(lootFrame,
        "Loot - " .. tostring(GetUnitName("target", false)),
        nil, "GameFontWhiteTiny")
    lootFrameTitle:SetPoint("LEFT", 10, 0)
    lootFrameTitle:SetPoint("TOP", 0, -7)
    self.frames.lootFrameTitle = lootFrameTitle

    local lootFrameScrollItems = {}
    local lootFrameScroll = self:CreateScrollFrame(lootFrame, 100, 20, 3,
        function(frame, index)

            local textureAndLabelWrapper = CreateFrame("Frame", nil, frame)
            textureAndLabelWrapper:SetSize(80, 80)
            textureAndLabelWrapper:SetPoint("LEFT", 25, 0)

            local texture = textureAndLabelWrapper:CreateTexture()
            texture:SetSize(40, 40)
            texture:SetPoint("CENTER", 0, 10)
            texture:SetTexture(1, 1, 1, 1)

            local label = self:CreateLabelStandard(textureAndLabelWrapper,
                "Not initialized",
                nil, "GameFontWhiteTiny")
            label:SetPoint("CENTER", 0, -25)
            label:SetWidth(80)

            -- parent, width, height, text
            local offerButton = LootSquid:CreateButtonStandard(frame, 150, 30, "Offer to Raid")
            offerButton:SetPoint("RIGHT", -20, 0)
            offerButton:SetPoint("TOP", 0, -20)

            local transferButton = LootSquid:CreateButtonStandard(frame, 150, 30, "Transfer to me")
            transferButton:SetPoint("RIGHT", -20, 0)
            transferButton:SetPoint("BOTTOM", 0, 20)

            lootFrameScrollItems[index] = {
                frame = frame,
                texture = texture,
                label = label,
                offerButton = offerButton,
                transferButton = transferButton,
                textureAndLabelWrapper = textureAndLabelWrapper
            }
        end);

    lootFrameScroll.items = lootFrameScrollItems

    self.frames.lootFrameScroll = lootFrameScroll;


    local overviewFrame = self:CreateFrameSolid(bodyFrame,
        (bodyFrame:GetWidth() - padding) / 2,
        (bodyFrame:GetHeight() - padding * 2) / 3,
        { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    overviewFrame:SetPoint("TOP", 0, 0)
    overviewFrame:SetPoint("RIGHT", 0, 0)

    local overviewLabel = self:CreateLabelStandard(overviewFrame,
        "Session",
        nil, "GameFontWhiteTiny")
    overviewLabel:SetPoint("LEFT", overviewFrame, "TOPLEFT", 10, -10)

    self.frames.overviewFrame = overviewFrame

    local overviewLeftLabel = self:CreateLabelStandard(overviewFrame,
        "overviewLeftLabel",
        nil, "GameFontWhiteTiny")
    overviewLeftLabel:SetRelativePoint("CENTER", "TOPLEFT", 0.22, -0.2)
    self.frames.overviewLeftLabel = overviewLeftLabel

    local overviewRightLabel = self:CreateLabelStandard(overviewFrame,
        "overviewRightLabel",
        nil, "GameFontWhiteTiny")
    overviewRightLabel:SetRelativePoint("CENTER", "TOPRIGHT", -0.22, -0.2)
    self.frames.overviewRightLabel = overviewRightLabel

    local overviewTextureAndLabelWrapper = LootSquid:CreateFrame("Frame", nil, overviewFrame)
    overviewTextureAndLabelWrapper:SetSize(80, 80)
    overviewTextureAndLabelWrapper:SetRelativePoint("CENTER", "TOP", 0, -0.2)
    self.frames.overviewTextureAndLabelWrapper = overviewTextureAndLabelWrapper

    local overviewTextureItem = overviewTextureAndLabelWrapper:CreateTexture()
    overviewTextureItem:SetSize(40, 40)
    overviewTextureItem:SetPoint("CENTER", 0, 10)
    overviewTextureItem:SetTexture(1, 1, 1, 1)
    self.frames.overviewTextureItem = overviewTextureItem

    local overviewLabelItem = LootSquid:CreateLabelStandard(overviewTextureAndLabelWrapper,
        "Not initialized",
        nil, "GameFontWhiteTiny")
    overviewLabelItem:SetPoint("CENTER", 0, -25)
    overviewLabelItem:SetWidth(80)
    self.frames.overviewLabelItem = overviewLabelItem


    self.frames.overviewCurrentItems = {}

    local currentItemsCnt = 6

    for i = 1, 6 do

        local xRel = 0.1 + (i-1) * (1-0.2) / (currentItemsCnt-1)

        local textureAndLabelWrapper = LootSquid:CreateFrame("Frame", nil, overviewFrame)
        textureAndLabelWrapper:SetSize(80, 80)
        textureAndLabelWrapper:SetRelativePoint("CENTER", "LEFT", xRel, -0.05)

        local textureItem = textureAndLabelWrapper:CreateTexture()
        textureItem:SetSize(40, 40)
        textureItem:SetPoint("CENTER", 0, 10)
        textureItem:SetTexture(1, 1, 1, 1)

        local labelItem = LootSquid:CreateLabelStandard(textureAndLabelWrapper,
            "Not initialized",
            nil, "GameFontWhiteTiny")
        labelItem:SetPoint("CENTER", 0, -25)
        labelItem:SetWidth(80)


        local needButton = LootSquid:CreateButtonStandard(overviewFrame, 85, 26, "Need")
        needButton:SetRelativePoint("CENTER", "LEFT", xRel, -0.28)

        local greedButton = LootSquid:CreateButtonStandard(overviewFrame, 85, 26, "Greed")
        greedButton:SetRelativePoint("CENTER", "LEFT", xRel, -0.4)

--        local playerNeedButton = LootSquid:CreateButtonStandard(overviewFrame, 150, 30, "Need it")
--        playerNeedButton:SetPoint("CENTER", overviewFrame, "TOPLEFT", 120, -110)
--        playerNeedButton:SetScript("OnClick",
--            function()
--                if (_G.LootSquid.session.currentLootSlot < 1) then
--                    LootSquid:LogError("No session is active. Cannot need an item.")
--                else
--                    _G.LootSquid.session.playersGreeding[UnitName("player")] = nil
--                    _G.LootSquid.session.playersNeeding[UnitName("player")] =
--                    { isNeeding = true, itemLinks = nil, comment = nil }
--                    LootSquid:UpdateRequestFrame()
--                end
--            end)
--
--        local playerGreedButton = LootSquid:CreateButtonStandard(overviewFrame, 150, 30, "Greed it")
--        playerGreedButton:SetPoint("CENTER", overviewFrame, "TOPLEFT", 120, -150)
--        playerGreedButton:SetScript("OnClick",
--            function()
--                if (_G.LootSquid.session.currentLootSlot < 1) then
--                    LootSquid:LogError("No session is active. Cannot greed an item.")
--                else
--                    _G.LootSquid.session.playersNeeding[UnitName("player")] = nil
--                    _G.LootSquid.session.playersGreeding[UnitName("player")] =
--                    { isGreeding = true, itemLinks = nil, comment = nil }
--                    LootSquid:UpdateRequestFrame()
--                end
--            end)


        self.frames.overviewCurrentItems[i] = {
            textureAndLabelWrapper = textureAndLabelWrapper,
            textureItem = textureItem,
            labelItem = labelItem,
        }
    end






    local requestFrame = self:CreateFrameSolid(bodyFrame,
        nil, --(bodyFrame:GetWidth()),
        nil, --(bodyFrame:GetHeight() - padding * 2) * 2 / 3 + padding,
        { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    requestFrame:SetPoint("TOPLEFT", 0, -(bodyFrame:GetHeight() - padding * 2) / 3 - padding)
    requestFrame:SetPoint("BOTTOMRIGHT", -16 - padding, 0)

    self.frames.requestFrame = requestFrame

    -- LootSquid:CreateFrame(frameType, name, parent, template)
    local requestHeaderFrame = LootSquid:CreateFrame("Frame", nil, requestFrame)

    requestHeaderFrame:SetRelativePoint("TOPLEFT", "TOPLEFT", 0.01, -0.005)
    requestHeaderFrame:SetRelativePoint("BOTTOMRIGHT", "TOPRIGHT", -0.01, -0.035)
    --requestHeaderFrame:SetBackdrop(LootSquid:CreateBackdropSolid(0))
    --requestHeaderFrame:SetBackdropColor(0.1, 0.1, 0.1, 1.0 )
    --requestHeaderFrame:SetBackdropBorderColor(edgeColor.r, edgeColor.g, edgeColor.b, edgeColor.a)

    local headerLabelOffsetY = -0.018

    local requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Player info",
        nil, "GameFontWhiteTiny")

    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.playerInfo, headerLabelOffsetY)

    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Current Item",
        nil, "GameFontWhiteTiny")

    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.currentItem, headerLabelOffsetY)

    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Improvement",
        nil, "GameFontWhiteTiny")

    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.improvement, headerLabelOffsetY)

    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Attendance",
        nil, "GameFontWhiteTiny")

    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.data1, headerLabelOffsetY)

    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Items Received",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.data2, headerLabelOffsetY)

    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "API",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.data3, headerLabelOffsetY)

    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Rating",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.data4, headerLabelOffsetY)

    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Request info",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.data5, headerLabelOffsetY)

    LootSquid.frames.requestHeaderLabelVotes = self:CreateLabelStandard(requestHeaderFrame,
        "Votes",
        nil, "GameFontWhiteTiny")
    LootSquid.frames.requestHeaderLabelVotes:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.votes, headerLabelOffsetY)


    requestHeaderLabel = self:CreateLabelStandard(requestHeaderFrame,
        "Actions",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetRelativePoint("CENTER", "LEFT",
        LootSquid.settings.layout.requestFrame.actions, headerLabelOffsetY)

    --print(LootSquid.settings.layout.requestFrame.actions)


    local requestFrameScrollItems = {}
    local requestFrameScroll = self:CreateScrollFrame(requestFrame, 100, 20, 3,
        function(frame, index)

            local labelName = self:CreateLabelStandard(frame,
                "labelName",
                nil, "GameFontWhiteTiny")
            labelName:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.playerInfo, 0.15)

            local labelSpec = self:CreateLabelStandard(frame,
                "labelSpec",
                nil, "GameFontWhiteTiny")
            labelSpec:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.playerInfo, -0.15)
            --
            --            local labelStatPrio = self:CreateLabelStandard(frame,
            --                "labelStatPrio",
            --                nil, "GameFontWhiteTiny")
            --            labelStatPrio:SetPoint("CENTER", frame, "BOTTOMLEFT", 80, 20)

            --[[
            currentItem = 0.16,  -- 1/1000
            improvement = 0.26,  -- 1/1000
            data1 = 0.36,        -- 1/1000
            data2 = 0.44,        -- 1/1000
            data3 = 0.52,        -- 1/1000
            data4 = 0.60,        -- 1/1000
            data5 = 0.68,        -- 1/1000
            votes = 0.8,
            actions = 0.94       -- 1/1000
            ]]


            local textureAndLabelWrapper = LootSquid:CreateFrame("Frame", nil, frame)
            textureAndLabelWrapper:SetSize(80, 80)
            textureAndLabelWrapper:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.currentItem, 0)

            local textureItem = textureAndLabelWrapper:CreateTexture()
            textureItem:SetSize(40, 40)
            textureItem:SetPoint("CENTER", 0, 10)
            textureItem:SetTexture(1, 1, 1, 1)

            local labelItem = LootSquid:CreateLabelStandard(textureAndLabelWrapper,
                "Not initialized",
                nil, "GameFontWhiteTiny")
            labelItem:SetPoint("CENTER", 0, -25)
            labelItem:SetWidth(80)

            local labelItemLvlIncrease = LootSquid:CreateLabelStandard(frame,
                "label ItemLvl Increase",
                nil, "GameFontWhiteTiny")
            labelItemLvlIncrease:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.improvement, 0)
            labelItemLvlIncrease:SetRelativeWidth(0.06)

            local labelData1 = LootSquid:CreateLabelStandard(frame,
                "labelData1",
                nil, "GameFontWhiteTiny")
            labelData1:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.data1, 0)

            local labelData2 = LootSquid:CreateLabelStandard(frame,
                "labelData2",
                nil, "GameFontWhiteTiny")
            labelData2:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.data2, 0)

            local labelData3 = LootSquid:CreateLabelStandard(frame,
                "labelData3",
                nil, "GameFontWhiteTiny")
            labelData3:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.data3, 0)

            local labelData4 = LootSquid:CreateLabelStandard(frame,
                "labelData4",
                nil, "GameFontWhiteTiny")
            labelData4:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.data4, 0)

            local labelRequestInfo = LootSquid:CreateLabelStandard(frame,
                "labelRequestInfo",
                nil, "GameFontWhiteTiny")
            labelRequestInfo:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.data5, 0.0)
            labelRequestInfo:SetRelativeWidth(0.06)

            local labelVotes = LootSquid:CreateLabelStandard(frame,
                "",
                nil, "GameFontWhiteTiny")
            labelVotes:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.votes, 0)


            local voteButton = LootSquid:CreateButtonStandard(frame, 125, 28, "Vote")
            voteButton:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.actions, 0.28)


            -- parent, width, height, text
            local awardButton = LootSquid:CreateButtonStandard(frame, 125, 28, "Award")
            awardButton:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.actions, 0.0)


            local rollButton = LootSquid:CreateButtonStandard(frame, 125, 28, "Whisper for roll")
            --rollButton:SetRelativeWidth(0.12)
            --rollButton:SetRelativeHeight(0.28)
            rollButton:SetRelativePoint("CENTER", "LEFT", LootSquid.settings.layout.requestFrame.actions, -0.28)

            requestFrameScrollItems[index] = {
                frame = frame,
                labelName = labelName,
                labelSpec = labelSpec,
                --labelStatPrio = labelStatPrio,
                textureAndLabelWrapper = textureAndLabelWrapper,
                textureItem = textureItem,
                labelItem = labelItem,
                labelData1 = labelData1,
                labelData2 = labelData2,
                labelData3 = labelData3,
                labelData4 = labelData4,
                labelVotes = labelVotes,
                voteButton = voteButton,
                labelRequestInfo = labelRequestInfo,
                awardButton = awardButton,
                rollButton = rollButton,
                labelItemLvlIncrease = labelItemLvlIncrease
            }
        end);

    requestFrameScroll.items = requestFrameScrollItems

    self.frames.requestFrameScroll = requestFrameScroll;




    self.frames.window = f
end

function LootSquid:UpdateGui()
end

function LootSquid:OpenLootGui()

    if self.frames.window == nil then
        --        print("pcall init")
        --        local success, error = pcall(LootSquid:InitGuiWrap)
        --        if not success then
        --            print("pcall error")
        --            LootSquid:LogError("InitGui failed: " .. tostring(error))
        --        else
        --            print("pcall success")
        --        end

        LootSquid:InitGui()
    end

    -- attempt to detect if ElvUI has loot frame override enabled
    --    for i = 0, 10 do
    --        if ElvUI
    --                and ElvUI[i]
    --                and ElvUI[i].Skins
    --                and ElvUI[i].Skins.db
    --                and ElvUI[i].Skins.db.blizzard
    --                and ElvUI[i].Skins.db.blizzard.loot == true
    --        then
    --            LootSquid:LogInfo("|cFFFF0000 Warning: ElvUI found and has loot frame override enabled. "
    --            .. "Please go to ElvUI config > General and untick 'loot'. "
    --            .. "Optionally also concider disabling loot frame skinning in Skins > Loot Frames. "
    --            .. "The issue stems from ElvUI placing the loot frame in very high frame strata (above everything else).")
    --        end
    --    end


    self.frames.window:Show()

    self:UpdateSessionFrame()


    local mode = LootSquid:GetMode()

    if mode == "MasterLooter" then
        LootSquid.frames.lootFrame:Show()
        LootSquid.frames.requestFrame:Show()
        --LootSquid.frames.overviewframe:SetRelativeWidth(1.0)
        --LootSquid.frames.overviewframe:SetRelativeHeight(1.0)


    elseif mode == "Council" then
        LootSquid.frames.lootFrame:Show()
        LootSquid.frames.requestFrame:Show()
       -- LootSquid.frames.overviewframe:SetRelativeWidth(1.0)
        --LootSquid.frames.overviewframe:SetRelativeHeight(1.0)
    else
        LootSquid.frames.window:SetRelativeWidth(LootSquid.settings.layout.windowFrame.width * 0.5)
        LootSquid.frames.window:SetRelativeHeight(LootSquid.settings.layout.windowFrame.height * 0.33)
        LootSquid.frames.window:UpdateLayout()
        LootSquid.frames.headerFrame:UpdateLayout()
        LootSquid.frames.bodyFrame:UpdateLayout()
        LootSquid.frames.footerFrame:UpdateLayout()

        LootSquid.frames.bodyFrame:Hide()
        LootSquid.frames.lootFrame:Hide()
        LootSquid.frames.requestFrame:Hide()
        LootSquid.frames.overviewFrame:Hide()
        LootSquid.frames.overviewFrame:SetRelativeWidth(1.0)
        LootSquid.frames.overviewFrame:SetRelativeHeight(1.0)
    end

   -- LootSquid.frames.bodyFrame:Hide()
    LootSquid.frames.window:UpdateLayout()
    LootSquid.frames.headerFrame:UpdateLayout()
    LootSquid.frames.bodyFrame:UpdateLayout()
    LootSquid.frames.footerFrame:UpdateLayout()






    if mode ~= "Player" then
        self:UpdateLootFrame()
        self:UpdateRequestFrame()

        if _G["LootSquidData"] ~= nil and _G["LootSquidData"].dataHash and _G["LootSquidData"].dataHashUnixTimeStamp then
            local utcDateStr = LootSquid:TimeToUtcDateStr(_G["LootSquidData"].dataHashUnixTimeStamp)

            local daysDiff = LootSquid:GetDateDifferenceTotalDays(
                LootSquid:TimeToDateTable(time()),
                LootSquid:TimeToDateTable(_G["LootSquidData"].dataHashUnixTimeStamp)
            )

            if daysDiff > 0.5 then
                LootSquid.frames.footerLabel:SetFormattedText("|cFFFFFFFFLootSquidData: "
                        .. string.format("|cFFFF0000%.1f", daysDiff) .. " days old|cFFFFFFFF, UTC Timestamp: "
                        .. utcDateStr .. ", Hash: " .. _G["LootSquidData"].dataHash)

                -- warn if old data and master looter
                if mode == "MasterLooter" and not LootSquid.data.hasWarned then
                    LootSquid.data.hasWarned = true
                    local popup = StaticPopup_Show("LOOTSQUID_WARNING_DATA_OLD")
                    if popup then
                        popup:SetFrameStrata("FULLSCREEN_DIALOG")
                    end
                end
            else
                LootSquid.frames.footerLabel:SetFormattedText("|cFFFFFFFFLootSquidData: "
                        .. string.format("|cFF00FF00%.1f", daysDiff) .. " days old|cFFFFFFFF, UTC Timestamp: "
                        .. utcDateStr .. ", Hash: " .. _G["LootSquidData"].dataHash)
            end
        else
            LootSquid.frames.footerLabel:SetFormattedText("|cFFFFFFFFLootSquidData: |cFFFF0000Not Loaded!")
        end
    else
        LootSquid.frames.footerLabel:SetFormattedText("")
    end


end

function LootSquid:OnUpdateLootTable()
    self:UpdateLootFrame()
end

function LootSquid:UpdateLootTable(doOpenGui)

    -- if player is the master looter
    if (_G.LootSquid.debug.alwaysParseLoot == true
            or (LootSquid:IsPlayerMasterLooter())) then

        for k in pairs(_G.LootSquid.session.loot) do
            _G.LootSquid.session.loot[k] = nil
        end

        local lootThreshold = GetLootThreshold()

        local lootSlots = GetNumLootItems()
        local lootSlotsAtleastThreshold = 0

        _G.LootSquid.session.lootSlots = lootSlots

        for lootSlot = 1, lootSlots do

            --[[
            -- The lootSlot will remain concistent for items as long as the loot box remains open.
            -- If the loot box is closed after some items have been looted, lootSlot will have changed
            -- for items which originally was placed after a looted item.
            -- ]]
            local itemTexture, itemName, itemQuantity, itemQuality, itemLocked = GetLootSlotInfo(lootSlot)

            if itemQuality and itemQuality >= lootThreshold then
                lootSlotsAtleastThreshold = lootSlotsAtleastThreshold + 1
            end

            local itemLootType = GetLootSlotType(lootSlot)


            --            self:LogInfo("GetLootSlotInfo "
            --                    .. tostring(lootSlot) .. ", "
            --                    .. tostring(itemName) .. ", "
            --                    .. tostring(itemQuantity) .. ", "
            --                    .. tostring(itemQuality) .. ", "
            --                    .. tostring(itemLocked) .. ", "
            --                    .. tostring(itemLootType)
            --            )

            local itemLink = GetLootSlotLink(lootSlot)

            _G.LootSquid.session.loot[lootSlot] = {
                itemName = itemName,
                itemQuantity = itemQuantity,
                itemQuality = itemQuality,
                itemLocked = itemLocked,
                itemLootType = itemLootType,
                itemLink = itemLink,
                candidates = {}
            }

            for candidateIndex = 1, 40 do

                local candidateName = GetMasterLootCandidate(lootSlot, candidateIndex)
                if candidateName then

                    _G.LootSquid.session.loot[lootSlot].candidates[candidateName] = {
                        candidateIndex = candidateIndex,
                    }

                else
                    LootSquid:LogDebug("no candidate name")
                end
            end
        end

        if ((_G.LootSquid.debug.alwaysParseLoot and doOpenGui)
                or
                (lootSlotsAtleastThreshold > 0
                        and doOpenGui
                        and lootSlots > 0)) then
            self:OpenLootGui()
        else
            self:OnUpdateLootTable()
        end
    end
end

function LootSquid:UpdateLootFrame()
    if (not self.frames.lootFrameScroll) then
        return
    end

    local lootFrameScrollItemsIndex = 0

    for k in pairs(self.session.loot) do

        local v = self.session.loot[k]

        --[[
        -- loot[lootSlot] = {
            itemName = itemName,
            itemQuantity = itemQuantity,
            itemQuality = itemQuality,
            itemLocked = itemLocked,
            itemLootType = itemLootType,
            itemLink = itemLink,
            candidates = {}
        }]]

        --[[
         lootFrameScrollItems[index] = {
            frame = frame,
            texture = texture,
            label = label,
            offerButton = offerButton,
            transferButton = transferButton
        }]]

        local lootThreshold = GetLootThreshold()

        if v and v.itemLink and v.itemQuality and v.itemQuality >= lootThreshold then

            local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice =
            GetItemInfo(v.itemLink)

            if self.frames.lootFrameScroll.items and
                    self.frames.lootFrameScroll.items[lootFrameScrollItemsIndex] then

                local lootFrameScrollItem = self.frames.lootFrameScroll.items[lootFrameScrollItemsIndex];

                lootFrameScrollItem.frame:Show()

                lootFrameScrollItem.textureAndLabelWrapper:EnableMouse(true)
                lootFrameScrollItem.textureAndLabelWrapper:SetScript("OnEnter", function(self, motion)
                    LootSquid:LootIconEnter(v.itemLink)
                end)
                lootFrameScrollItem.textureAndLabelWrapper:SetScript("OnLeave", function(self, motion)
                    LootSquid:LootIconLeave()
                end)

                lootFrameScrollItem.texture:SetTexture(texture)

                local r, g, b, hex = GetItemQualityColor(quality)

                lootFrameScrollItem.label:SetFormattedText("|c" .. hex .. name)

                lootFrameScrollItem.offerButton:RegisterForClicks("LeftButtonUp")
                lootFrameScrollItem.offerButton:SetScript("OnClick",
                    function(widget)
                        LootSquid:ClearNeedAndGreedRequests()
                        LootSquid.session.currentLootSlot = k
                        LootSquid.session.currentLootOffered = {
                            name = name,
                            link = link,
                            quality = quality,
                            iLevel = iLevel,
                            class = class,
                            subclass = subclass,
                            maxStack = maxStack,
                            equipSlot = equipSlot,
                            texture = texture
                        }

                        LootSquid:RaidWarning("LootSquid: Loot session starting for: " .. v.itemLink)
                        SendChatMessage("LootSquid: write either !need <currentItem> <comment> or !greed <currentItem> <comment> in chat. If you change your mind: write !pass",
                            "RAID")
                    end)

                lootFrameScrollItem.transferButton:RegisterForClicks("LeftButtonUp")
                lootFrameScrollItem.transferButton:SetScript("OnClick",
                    function(widget)
                        LootSquid.session.currentLootSlot = k
                        StaticPopupDialogs["LOOTSQUID_AWARD_LOOT"].text = "Really transfer item to you?"
                        local popup = StaticPopup_Show("LOOTSQUID_AWARD_LOOT", UnitName("player"))
                        if popup then
                            popup:SetFrameStrata("FULLSCREEN_DIALOG")
                            popup.playerName = UnitName("player")
                        end
                    end)
            end

            lootFrameScrollItemsIndex = lootFrameScrollItemsIndex + 1
        end
    end


    if lootFrameScrollItemsIndex < 1 then
        self.frames.lootFrameScroll.frameMessage:SetText("No more items available.")
        self.frames.lootFrameScroll.frameMessage:Show()
    else
        self.frames.lootFrameScroll.frameMessage:Hide()
    end

    LootSquid.frames.lootFrameScroll:HideIndexAndAfter(self.frames.lootFrameScroll, lootFrameScrollItemsIndex)
end

function LootSquid:UpdateSessionFrame()
    --[[LootSquid.session.currentLootOffered = {
        name = name,
        link = link,
        quality = quality,
        iLevel = iLevel,
        class = class,
        subclass = subclass,
        maxStack = maxStack,
        equipSlot = equipSlot,
        texture = texture
    }]]

    if not LootSquid.session or not LootSquid.session.currentLootOffered then
        return
    end

    local loot = LootSquid.session.currentLootOffered

    self.frames.overviewTextureAndLabelWrapper:EnableMouse(true)
    self.frames.overviewTextureAndLabelWrapper:SetScript(
        "OnEnter",
        function(self, motion)
            LootSquid:LootIconEnter(loot.link)
        end)

    self.frames.overviewTextureAndLabelWrapper:SetScript("OnLeave",
        function(self, motion)
            LootSquid:LootIconLeave()
        end)

    local r, g, b, hex = GetItemQualityColor(loot.quality)
    self.frames.overviewLabelItem:SetFormattedText("|c" .. hex .. loot.name)

    self.frames.overviewTextureItem:SetTexture(loot.texture)
end

function LootSquid:GetStandingsByPlayerName(playerName)

    --    local playerRealmName = GetRealmName()
    --    local realmIndex = string.find(playerName, "-"..tostring(playerRealmName))
    --
    --    local playerNameFixed = playerName
    --    if realmIndex then
    --        playerNameFixed = string.sub(playerName, 1, realmIndex-1)
    --        LootSquid:LogDebug("Fixing player name due to local realm to: "..tostring(playerName))
    --    end

    local playerNameFixed = playerName
    local dashIndex = string.find(playerName, "-")
    if dashIndex then
        -- http://wow.gamepedia.com/API_Ambiguate
        playerNameFixed = Ambiguate(playerName, "none")
        LootSquid:LogDebug("Fixing player name due to local realm from: " .. tostring(playerName) .. " to :" .. tostring(playerNameFixed))
    end

    LootSquid:LogDebug("Standings lookup for player: " .. tostring(playerNameFixed) .. ", dashIndex: " .. tostring(dashIndex))

    if (_G["LootSquidData"] and _G["LootSquidData"].data and _G["LootSquidData"].data[playerNameFixed]) then
        return _G["LootSquidData"].data[playerNameFixed], playerNameFixed
    end
    return nil
end

function LootSquid:SortStandingsByApiRating(a, b)

    -- todo: support sorting by other data fields
    local a = LootSquid:GetStandingsByPlayerName(a)
    local b = LootSquid:GetStandingsByPlayerName(b)

    if (a == nil or a.data4 == nil) and (b == nil or b.data4 == nil) then
        return false
    end
    if (a == nil or a.data4 == nil) then
        return false
    end
    if (b == nil or b.data4 == nil) then
        return true
    end

    return a.data4 > b.data4
end

function LootSquid:AddStandingsDataRows(requestFrameScrollItemsIndex, needingOrGreedingPlayers, isNeed)

    if (not self.frames.requestFrameScroll) then
        return -1
    end

    local mode = LootSquid:GetMode()

    local totalVotes = 0

    for k, v in self:spairs(needingOrGreedingPlayers,
        function(t, a, b) return self:SortStandingsByApiRating(a, b) end) do

        local data, playerNameFixed = LootSquid:GetStandingsByPlayerName(k)

        local data1 = 0
        local data1p = 0
        local data2 = 0
        local data3 = 0
        local data4 = 0

        if (data and data.data1 and data.data1p and data.data2 and data.data3 and data.data4) then
            data1 = data.data1
            data1p = data.data1p
            data2 = data.data2
            data3 = data.data3
            data4 = data.data4
        else
            LootSquid:LogDebug("No standings for player name: " .. tostring(k))
        end

        --[[
        --
        -- requestFrameScrollItems[index] = {
                frame = frame,
                labelName = labelName,
                labelSpec = labelSpec,
                labelStatPrio = labelStatPrio,
                textureItem = textureItem,
                labelItem = labelItem,
                labelItems12 = labelItems12,
                labelAttendance12 = labelAttendance12,
                labelApi = labelApi,
                labelNeedGreed = labelNeedGreed,
                awardButton = awardButton,
                rollButton = rollButton
            }
        --
        -- ]]


        if self.frames.requestFrameScroll.items and
                self.frames.requestFrameScroll.items[requestFrameScrollItemsIndex] then

            local requestFrameScrollItem = self.frames.requestFrameScroll.items[requestFrameScrollItemsIndex];

            local ambiguatedName = Ambiguate(k, "none")
            local class, classFileName = UnitClass(ambiguatedName)
            local color = RAID_CLASS_COLORS[classFileName]
            local playerNameColored
            if color then
                playerNameColored = string.format("|cFF%02X%02X%02X %s", color.r * 255, color.g * 255, color.b * 255, ambiguatedName)
            else
                playerNameColored = string.format("|cFF%02X%02X%02X %s", 0.5 * 255, 0.5 * 255, 0.5 * 255, ambiguatedName)
            end

            requestFrameScrollItem.frame:Show()
            requestFrameScrollItem.labelName:SetFormattedText(playerNameColored)


            requestFrameScrollItem.labelSpec:SetText(class)



            local standingColor = "FF0000"

            if data1p >= 87.5 then -- miss up to 1 of 8 raids and still be green
            standingColor = "00FF00" -- green
            requestFrameScrollItem.labelData1:SetFont("Fonts\\FRIZQT__.TTF", 12)
            elseif data1 >= 66.7 then -- mythic raiding and only 2/3 attendance, really?
            --standingColor = "9999FF" -- blue
            standingColor = "FFFFFF" -- blue
            requestFrameScrollItem.labelData1:SetFont("Fonts\\FRIZQT__.TTF", 11)
            else
                requestFrameScrollItem.labelData1:SetFont("Fonts\\FRIZQT__.TTF", 10)
            end

            requestFrameScrollItem.labelData1:SetFormattedText(string.format("|cFF%s %s (%.2f %s)", standingColor, string.format("%.2f", data1), data1p, "%%"))
            requestFrameScrollItem.labelData2:SetText(string.format("%.2f", data2))
            requestFrameScrollItem.labelData3:SetText(string.format("%.2f", data3))
            requestFrameScrollItem.labelData4:SetText(string.format("%.2f", data4))

            --[[
            -- v holds:
            -- comment = comment, itemLinks = itemLinks
            -- ]]

            if v.itemLinks and LootSquid:tablelength(v.itemLinks) > 0 and v.itemLinks[1] then
                -- request the item info, if we already have it cached.
                -- An immidiate call after will either be nil (and pending server request) or contain the info.
                local currentItemLinkFromMessage = v.itemLinks[1]

                local a, b, Color, Ltype, zId, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, zName =
                string.find(currentItemLinkFromMessage,
                    "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*):?(%-?%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")

                GetItemInfo(currentItemLinkFromMessage)

                local currentItemName,
                currentItemLink,
                currentItemQuality,
                currentItemLevel, currentItemMinLevel, currentItemClass, currentItemSubclass,
                currentItemMaxStack, currentItemEquipSlot, currentItemTexture, currentItemVendorPrice =
                GetItemInfo(currentItemLinkFromMessage)

                if (currentItemName --and currentItemLink
                        and currentItemQuality and currentItemTexture
                        and currentItemLevel and currentItemEquipSlot) then

                    requestFrameScrollItem.textureAndLabelWrapper:EnableMouse(true)
                    requestFrameScrollItem.textureAndLabelWrapper:SetScript("OnEnter",
                        function(self, motion)
                            --[[ {
                                isNeeding = true,
                                currentItemId = Id
                                ]]

                            --if (currentItemLink) then
                            LootSquid:LootIconEnter(currentItemLinkFromMessage)
                            --end
                        end)

                    requestFrameScrollItem.textureAndLabelWrapper:SetScript("OnLeave",
                        function(self, motion)
                            LootSquid:LootIconLeave()
                        end)

                    local r, g, b, hex = GetItemQualityColor(currentItemQuality)
                    requestFrameScrollItem.labelItem:SetFormattedText("|c" .. hex .. currentItemName)


                    --LootSquid:LogDebug("Setting current item texture: " .. currentItemTexture)
                    requestFrameScrollItem.textureItem:SetTexture(currentItemTexture)


                    --[[
                        LootSquid.session.currentLootOffered = {
                            name = name,
                            link = link,
                            quality = quality,
                            lLevel = iLevel,
                            class = class,
                            subclass = subclass,
                            maxStack = maxStack,
                            equipSlot = equipSlot,
                            texture = texture
                        }
                    -- ]]

                    if LootSquid.session.currentLootOffered then

                        local iLevelDiff = (LootSquid.session.currentLootOffered.iLevel
                                and LootSquid.session.currentLootOffered.iLevel or 0) - currentItemLevel

                        if LootSquid.session.currentLootOffered.equipSlot ~= currentItemEquipSlot then
                            requestFrameScrollItem.labelItemLvlIncrease:SetFormattedText("|cFFFF0000SLOT MISMATCH! "
                                    .. tostring(LootSquid.session.currentLootOffered.equipSlot)
                                    .. " vs " .. tostring(currentItemEquipSlot) .. ", diff: " .. tostring(iLevelDiff))
                        else

                            local prefixColor
                            if iLevelDiff < 0 then
                                prefixColor = "|cFFFF0000"
                            elseif iLevelDiff == 0 then
                                prefixColor = "|cFFFFFFFF"
                            else
                                prefixColor = "|cFF00FF00"
                            end

                            requestFrameScrollItem.labelItemLvlIncrease:SetFormattedText(prefixColor .. tostring(iLevelDiff))
                        end
                    end


                else
                    requestFrameScrollItem.textureAndLabelWrapper:EnableMouse(false)
                    requestFrameScrollItem.textureItem:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark")
                    requestFrameScrollItem.labelItem:SetText("Item info pending from server...")
                    requestFrameScrollItem.labelItemLvlIncrease:SetText("?")
                end



            else
                requestFrameScrollItem.textureAndLabelWrapper:EnableMouse(false)
                requestFrameScrollItem.textureItem:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark")
                requestFrameScrollItem.labelItem:SetText("No item linked")
                requestFrameScrollItem.labelItemLvlIncrease:SetText("?")
            end


            local labelRequestInfoText = ""
            if isNeed then
                labelRequestInfoText = labelRequestInfoText .. "|cFF00FF00Item was needed"
            else
                labelRequestInfoText = labelRequestInfoText .. "|cFFFFFF00Item was greeded"
            end

            if v.comment and string.len(tostring(v.comment)) > 0 then
                labelRequestInfoText = labelRequestInfoText .. "\n\n|cFFFFFFFF" .. tostring(v.comment)
            end

            requestFrameScrollItem.labelRequestInfo:SetFormattedText(labelRequestInfoText)


            --labelVotes,
            --voteButton

            local votes = LootSquid.voting.votes[ambiguatedName]

            if votes then
                local voterNames = ""
                for voterId, voterName in pairs(votes.votedBy) do
                    voterNames = voterNames .. voterName .. "\n"
                    totalVotes = totalVotes + 1
                end
                requestFrameScrollItem.labelVotes:SetFormattedText(voterNames)
            else
                requestFrameScrollItem.labelVotes:SetFormattedText("")
            end

            requestFrameScrollItem.voteButton:SetScript("OnClick",
                function(widget)
                    LootSquid:SendVote(ambiguatedName)
                end)


            if mode == "MasterLooter" then

                requestFrameScrollItem.awardButton:RegisterForClicks("LeftButtonUp")
                requestFrameScrollItem.awardButton:SetScript("OnClick",
                    function(widget)
                        StaticPopupDialogs["LOOTSQUID_AWARD_LOOT"].text = "Really transfer item to: " .. k
                        local popup = StaticPopup_Show("LOOTSQUID_AWARD_LOOT", k)
                        if popup then
                            popup:SetFrameStrata("FULLSCREEN_DIALOG")
                            popup.playerName = k
                        end
                    end)
                requestFrameScrollItem.awardButton:Enable()

                requestFrameScrollItem.rollButton:RegisterForClicks("LeftButtonUp")
                requestFrameScrollItem.rollButton:SetScript("OnClick",
                    function(widget)
                        LootSquid:WhisperToPlayer(k,
                            "The loot master has requested that you please perform a random roll as part of the loot distribution.")
                    end)
                requestFrameScrollItem.rollButton:Enable()
            else
                requestFrameScrollItem.awardButton:Disable()
                requestFrameScrollItem.rollButton:Disable()
            end
        end

        requestFrameScrollItemsIndex = requestFrameScrollItemsIndex + 1
    end

    return requestFrameScrollItemsIndex, totalVotes
end

function LootSquid:WhisperToPlayer(playerName, message)
    if LootSquid.debug.disableWhispers then
        return
    end

    SendChatMessage("LootSquid: " .. tostring(message),
        "WHISPER",
        nil,
        playerName)
end

function LootSquid:RaidWarning(message)
    SendChatMessage("LootSquid: " .. tostring(message),
        "RAID_WARNING")
end

function LootSquid:UpdateRequestFrame()

    --LootSquid.lastRequestFrameUpdatedTime = GetTime()

    if not LootSquid:IsActive() then
        return
    end

    LootSquid:LogDebug("UpdateRequestFrame")

    local offset, totalVotesNeed = self:AddStandingsDataRows(0, self.session.playersNeeding, true)

    if offset < 0 then
        self:LogError("offset was below zero for session.playersNeeding, this is not supposed to happen! Bailing out.")
        return
    end

    local totalVotesGreed

    offset, totalVotesGreed = self:AddStandingsDataRows(offset, self.session.playersGreeding, false)

    local totalVotes = totalVotesNeed + totalVotesGreed

    if totalVotes > 0 then
        LootSquid.frames.requestHeaderLabelVotes:SetFormattedText("Votes |cFFFFFF00(" .. tostring(totalVotes) .. ")")
    else
        LootSquid.frames.requestHeaderLabelVotes:SetFormattedText("Votes |cFFFFFFFF(none)")
    end


    if offset < 1 then
        self.frames.requestFrameScroll.frameMessage:SetText("No requests have been received yet.")
        self.frames.requestFrameScroll.frameMessage:Show()
    else
        self.frames.requestFrameScroll.frameMessage:Hide()
    end

    LootSquid.frames.requestFrameScroll:HideIndexAndAfter(self.frames.requestFrameScroll, offset)
end

function LootSquid:SendVote(playerName)
    LootSquid:LogDebug("Voting for player: " .. tostring(playerName))

    if (not LootSquid.session
            or not LootSquid.session.currentLootOffered
            or not LootSquid.session.id) then
        LootSquid:LogError("No loot session active, cannot vote.")
        return
    end

    local data

    if LootSquid:IsPlayerMasterLooter() then

        data = LootSquid:Serialize("VOTE-REQUEST",
            playerName,
            LootSquid.session)
    else
        data = LootSquid:Serialize("VOTE-SUBMIT",
            playerName)
    end


    LootSquid:LogDebug("Sending vote to officer channel")
    LootSquid:SendCommMessage("LootSquid", data, "OFFICER")
    LootSquid:LogDebug("Done sending vote to officer channel")

    if LootSquid.debug.addonComm then
        LootSquid:LogDebug("addon comm enabled")
        for k, v in pairs(LootSquid.debug.addonCommDebugee) do
            if v then
                local addonCommDebugee = Ambiguate(v, "none")
                if UnitIsConnected(addonCommDebugee) then
                    LootSquid:SendCommMessage("LootSquid", data, "WHISPER", addonCommDebugee)
                end
            end
        end
    end
end

function LootSquid:ClearNeedAndGreedRequests()
    self.session.playersNeeding = {}
    self.session.playersGreeding = {}
    LootSquid:UpdateRequestFrame()
    _G.LootSquid.session.currentLootSlot = 0

    if self.frames.overviewCurrentSessionItemName then
        self.frames.overviewCurrentSessionItemName:SetText("No session is active")
    end
end

function LootSquid:IsActive()
    if LootSquid.frames.window and LootSquid.frames.window:IsVisible() then
        return true
    end

    return false
end