function LootSquid:AddEvent(event, func)
    self.eventHandlers[event] = func
end

function LootSquid:RegisterEvents()

    self:LogDebug("RegisterEvents()")

    -- hook chat events
    self.eventHandlers["CHAT_MSG_RAID"] = self.OnChatMessage
    self.eventHandlers["CHAT_MSG_RAID_LEADER"] = self.OnChatMessage
    self.eventHandlers["CHAT_MSG_WHISPER"] = self.OnChatMessage
    --self.eventHandlers["CHAT_MSG_OFFICER"] = self.OnVoteChatMessage
    self.eventHandlers["LOOT_OPENED"] = self.OnLootOpened
    self.eventHandlers["LOOT_CLOSED"] = self.OnLootClosed
    self.eventHandlers["LOOT_SLOT_CHANGED"] = self.OnLootSlotChanged
    self.eventHandlers["LOOT_SLOT_CLEARED"] = self.OnLootSlotCleared
    self.eventHandlers["GET_ITEM_INFO_RECEIVED"] = self.OnGetItemInfoReceived
   -- self.eventHandlers["CHAT_MSG_ADDON"] = self.OnAddonMessageReceived

        -- register all event handlers on the main frame
    for k in pairs(self.eventHandlers) do
        self.frames.mainFrame:RegisterEvent(k)
    end

    -- add on OnEvent handler on the main frame
    self.frames.mainFrame:SetScript("OnEvent",
        function(mainFrame, event, ...)
        --self:LogInfo("OnEvent: " .. event)
            self.eventHandlers[event](self, ...)
        end)

    self.frames.mainFrame:SetScript("OnUpdate", LootSquid.OnUpdate)

    local AceComm = LibStub("AceComm-3.0")
    AceComm:Embed(LootSquid)
    AceComm:RegisterComm("LootSquid", LootSquid.OnAceCommReceived)

    local AceSerializer = LibStub("AceSerializer-3.0")
    AceSerializer:Embed(LootSquid)

    local AceTimer = LibStub("AceTimer-3.0")
    AceTimer:Embed(LootSquid)
end

function LootSquid:OnUpdate()

    if not LootSquid:IsActive() then
        return
    end

    --[[
    lastChatRequestEventTime = 0,
    lastGetItemInfoEventTime = 0,
    lastRequestFrameUpdatedTime = 0,
    -- ]]

    --print(tostring(LootSquid.lastChatRequestEventTime) .. " vs " .. tostring(LootSquid.lastRequestFrameUpdatedTime))

    local now = GetTime()

    if LootSquid.lastChatRequestEventTime > LootSquid.lastRequestFrameUpdatedTime
        or LootSquid.lastGetItemInfoEventTime > LootSquid.lastRequestFrameUpdatedTime then

        -- request update
        LootSquid:LogDebug("Requesting update")
        LootSquid.lastRequestFrameUpdateRequestTime = now
        LootSquid.lastChatRequestEventTime = 0
        LootSquid.lastGetItemInfoEventTime = 0
    else
        --LootSquid:LogDebug("not request for update")
    end

    local triggerDelaySecs = 1.0
    local triggerWindowSecs = 5.0

    if LootSquid.lastRequestFrameUpdateRequestTime
            and (now - LootSquid.lastRequestFrameUpdateRequestTime) > triggerDelaySecs
            and (now - LootSquid.lastRequestFrameUpdateRequestTime) < (triggerDelaySecs + triggerWindowSecs) then

        LootSquid:LogDebug("Updating request frame because 3 seconds have passed.")

        LootSquid.lastRequestFrameUpdateRequestTime = 0
        LootSquid.lastRequestFrameUpdatedTime = now
        LootSquid:UpdateRequestFrame()
    end
end

function LootSquid:OnGetItemInfoReceived(a, b, c, d, e, f, g, h)

    if not LootSquid:IsActive() then
        return
    end

    LootSquid:LogDebug("OnGetItemInfoReceived: "
            .. tostring(a) .. ", " .. tostring(b) .. ", " .. tostring(c) .. ", " .. tostring(d) .. ", "
            .. tostring(e) .. ", " .. tostring(f) .. ", " .. tostring(g) .. ", " .. tostring(h) .. ", ")

    LootSquid.lastGetItemInfoEventTime = GetTime()
end

function LootSquid:OnLootSlotChanged(a, b, c, d, e, f, g, h)
    self:LogDebug("OnLootSlotChanged()")
    --    LootSquid:LogInfo("OnLootSlotChanged: "
    --            ..tostring(a)..", "..tostring(b)..", "..tostring(c)..", "..tostring(d)..", "
    --            ..tostring(e)..", "..tostring(f)..", "..tostring(g)..", "..tostring(h)..", ")
end

--[[
-- Fires when the contents of a loot slot are removed.
-- Can fire due to the player looting the slot's contents
-- or due to them being taken by another group member.
 ]]
function LootSquid:OnLootSlotCleared(event, slotId)

    if not LootSquid:IsActive() then
        return
    end

    self:LogDebug("OnLootSlotCleared. slotId: " .. tostring(slotId))
    -- slotId: The numeric id of the slot that was looted.
    LootSquid:UpdateLootTable(false)
end

function LootSquid:OnChatMessage(message, sender, language, channelString, target, flags,
unknown, channelNumber, channelName, unknown, counter, guid)

    if not LootSquid:IsActive() then
        return
    end

    if LootSquid.debug.bypassSafeCalls then
        LootSquid:HandleOnChatMessage(message, sender)
    else
        local success, error = pcall(LootSquid.HandleOnChatMessage, nil, message, sender)
        if not success then
            LootSquid:LogError("OnChatMessage failed: " .. tostring(error))
        end
    end
end

function LootSquid:HandleOnChatMessage(message, sender)

    if not message or not sender then
        return
    end

    LootSquid:LogDebug("OnChatMessage. Sender: " .. tostring(sender) .. ", message: " .. tostring(message))

    -- do not process chat message unless master looter
    if not (LootSquid.debug.alwaysParseChat or LootSquid:IsPlayerMasterLooter()) then
        return
    end

    --local a, b, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, Name =
    --string.find(message,
    --    "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*):?(%-?%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")

    local comment = message
    comment = string.gsub(comment, "!need", "")
    comment = string.gsub(comment, "!greed", "")
    comment = string.gsub(comment, "!pass", "")


    local itemLinkMatch = "|%x+|Hitem:.-|h.-|h|r"
    local itemLinks = {}

    local maxIter = 5

    while true do

        maxIter = maxIter - 1

        local i, j = string.find(comment, itemLinkMatch)

        if maxIter < 1 or not i then
            break
        end

        local itemLink = string.sub(comment, i, j)
        table.insert(itemLinks, itemLink)

        local commentStripped = ""

        if i > 1 then
            commentStripped = commentStripped .. string.sub(comment, 1, i - 1)
        end

        if j < string.len(comment) then
            commentStripped = commentStripped .. string.sub(comment, j + 1)
        end

        comment = commentStripped
    end

    comment = comment:gsub("^%s*(.-)%s*$", "%1")

    local isNeed = LootSquid:strStartsWith(string.lower(message), "!need")
    local isGreed = LootSquid:strStartsWith(string.lower(message), "!greed")
    local isPass = LootSquid:strStartsWith(string.lower(message), "!pass")

    if (_G.LootSquid.session.currentLootSlot < 1 and (isNeed or isGreed or isPass)) then
        LootSquid:WhisperToPlayer(sender, "The loot master does not currently have an active loot session")
        return
    end

    if (isNeed) then

        _G.LootSquid.session.playersNeeding[sender] = { isNeeding = true, comment = comment, itemLinks = itemLinks }
        _G.LootSquid.session.playersGreeding[sender] = nil
        LootSquid:WhisperToPlayer(sender, "Your need request have been received. Please wait.")
    end

    if (isGreed) then

        _G.LootSquid.session.playersGreeding[sender] = { isGreeding = true, comment = comment, itemLinks = itemLinks }
        _G.LootSquid.session.playersNeeding[sender] = nil
        LootSquid:WhisperToPlayer(sender, "Your greed request have been received. Please wait.")
    end

    if (isPass) then

        _G.LootSquid.session.playersGreeding[sender] = nil
        _G.LootSquid.session.playersNeeding[sender] = nil
        LootSquid:WhisperToPlayer(sender, "Your pass request have been received. You have been removed from concideration.")
    end

    LootSquid.lastChatRequestEventTime = GetTime()
end

function LootSquid:OnVoteChatMessage(message, sender)

    if LootSquid.debug.bypassSafeCalls then
        LootSquid:HandleOnVoteChatMessage(message, sender)
    else
        local success, error = pcall(LootSquid.HandleOnVoteChatMessage, nil, message, sender)
        if not success then
            --LootSquid:LogError("OnVoteChatMessage failed: " .. tostring(error))
        end
    end
end

function LootSquid:HandleOnVoteChatMessage(message, sender)
    --LootSquid:LogDebug("HandleOnVoteChatMessage: "..tostring(message)..", "..tostring(sender))

end

--[[function LootSquid:DeserializeAddonMessage(message)

    -- version request
    if LootSquid:strStartsWith(message, "VERR_") then

    end

    -- version notification
    if LootSquid:strStartsWith(message, "VERN_")
            and not LootSquid.versionNewNotificationTriggered then
        LootSquid.versionNewNotificationTriggered = true
        print("Version available: "..message)
    end

    -- vote request
    if LootSquid:strStartsWith(message, "VOTR_") then

    end

    -- vote notify
    if LootSquid:strStartsWith(message, "VOTN_") then

    end
end

function LootSquid:OnAddonMessageReceived(prefix, message, channel, sender, a, b, c)
    ]]--[[print(tostring(prefix) .. ": "..tostring(message)
            .. " received from: " .. tostring(sender).." via " ..tostring(channel)
            .. ", a:" .. tostring(a)
            .. ", b:" .. tostring(b)
            .. ", c:" .. tostring(c)
    )]]--[[

    if prefix and message and channel and sender and prefix == "LootSquid" then

        local messageInfo = LootSquid:DeserializeAddonMessage(message)


    end

end]]

function LootSquid:AddVote(playerName, votedBy)

    -- remove any existing votes by voter
    for k, v in pairs(LootSquid.voting.votes) do
        for k2, v2 in pairs(v.votedBy) do
           if v2 == votedBy then
               v.votedBy[k2] = nil
           end
        end
    end

    if not LootSquid.voting.votes[playerName] then
        LootSquid.voting.votes[playerName] = { votedBy = { } }
    end

    table.insert(LootSquid.voting.votes[playerName].votedBy, votedBy)
end

function LootSquid:OnAceCommReceived(message, distribution, sender, a, b)

    if not message or not distribution or not sender then
        return
    end

    LootSquid:LogDebug("OnAceCommReceived: "
            ..tostring(message)..", "
            ..tostring(distribution)..", "
            ..tostring(sender)..", "
            ..tostring(a)..", "
            ..tostring(b)
    )

    local success, commandOrError = LootSquid:Deserialize(message)

    if not success then
        LootSquid:LogDebug("Failed to deserialize message: " .. tostring(commandOrError))
    end

    local command = commandOrError


    local isMessageFromPlayer = Ambiguate(sender, "none") == Ambiguate(UnitName("player"), "none")


    if command == "VOTE-REQUEST" then

        local _, _, playerName, lootSession = LootSquid:Deserialize(message)

        if not isMessageFromPlayer and lootSession then
            LootSquid.session = lootSession
            -- TODO: call update session frame
        end

        LootSquid:AddVote(playerName, sender)
        LootSquid:UpdateRequestFrame()

    elseif command == "VOTE-SUBMIT" then
        local _, _, playerName = LootSquid:Deserialize(message)

        LootSquid:AddVote(playerName, sender)
        LootSquid:UpdateRequestFrame()

    elseif command == "VERSION-REQUEST" and not isMessageFromPlayer then
        LootSquid:LogDebug("OnAceCommReceived: received version request. Sending version")
        -- SendCommMessage(prefix, text, distribution, target, prio, callbackFn, callbackArg)

        LootSquid:ScheduleTimer(function()

        LootSquid:SendCommMessage(
            "LootSquid",
            LootSquid:Serialize(
                "VERSION-RESPONSE",
                LootSquid.versionMajor,
                LootSquid.versionMinor,
                LootSquid.versionRevision,
                LootSquid.versionState),
            "WHISPER",
            sender,
            nil,
            function(arg, bytesSent, bytesTotal)
                LootSquid:LogDebug("OnAceCommReceived: callback send version response. "
                        ..tostring(arg).. ", " ..tostring(bytesSent).. " of " .. tostring(bytesTotal))
            end
        )
        end, math.random() * 3.0 + 2.0)

        LootSquid:LogDebug("OnAceCommReceived: send version response")
    elseif command == "VERSION-RESPONSE" then

    end

end

function LootSquid:OnLootOpened(autoLoot)
    LootSquid:LogDebug("OnLootOpened. autoLoot: " .. tostring(autoLoot))
    LootSquid:UpdateLootTable(LootSquid:IsPlayerMasterLooter())
end

function LootSquid:OnLootClosed()
    if LootSquid.frames.window and LootSquid.frames.window:IsVisible() then
       LootSquid:LogError("Loot window closed before loot squid window. Please do not move while processing loot.")
       LootSquid.frames.window:Hide()
    end
end

LootSquid.components.events = true