LootSquid.frames.autoIdCnt = 0

function LootSquid:PatchUpdateLayout(f)
    f.UpdateLayout = function()

        if f.isRelativeSizesPatched then

            local relativeWidth = f:GetRelativeWidth()
            local relativeHeight = f:GetRelativeHeight()

            if relativeWidth then
                f:SetRelativeWidth(relativeWidth)
            end

            if relativeHeight then
                f:SetRelativeHeight(relativeHeight)
            end
        end

        if f.isRelativeAnchorsPatched then
            f:UpdateRelativePoints()
        end
    end
end

function LootSquid:PatchRelativeAnchors(f)

    if f and f.isRelativeAnchorsPatched == true then
        return
    end

    -- labelItem:SetPoint("CENTER", 0, -25)
    -- labelItemLvlIncrease:SetPoint("CENTER", frame, "LEFT", 350, 0)
    f.SetRelativePoint = function(self, point, parentPoint, relativeX, relativeY)

        --print("x: "..tostring(relativeX * f:GetParent():GetWidth()))
        --print("y: "..tostring(relativeY * f:GetParent():GetHeight()))

        table.insert(f.relativePoints,
            {
                point = point,
                parentPoint = parentPoint,
                relativeX = relativeX,
                relativeY = relativeY
            })


        -- HACK: implicitly update
        f.UpdateRelativePoints()
    end

    f.UpdateRelativePoints = function()
        if not f.relativePoints then
            return
        end

        for k, v in pairs(f.relativePoints) do
            f:SetPoint(v.point, f:GetParent(), v.parentPoint,
                v.relativeX <= 1.0 and v.relativeX * f:GetParent():GetWidth() or v.relativeX,
                v.relativeY <= 1.0 and v.relativeY * f:GetParent():GetHeight() or v.relativeY)
        end
    end

    f.ClearRelativePoints = function()
        f.relativePoints = {}
    end

    f.relativePoints = {}
    f.isRelativeAnchorsPatched = true
end

function LootSquid:PatchRelativeSizing(f)

    if f and f.isRelativeSizesPatched == true then
        return
    end

    f.SetRelativeWidth = function(self, relativeWidth)
        f.relativeWidth = relativeWidth
        --print(f:GetParent():GetWidth())
        --print(relativeWidth)
        f:SetWidth(f:GetParent():GetWidth() * relativeWidth)
    end
    f.SetRelativeHeight = function(self, relativeHeight)
        f.relativeHeight = relativeHeight
        --print("relativeHeight: "..tostring(relativeHeight))
        f:SetHeight(f:GetParent():GetHeight() * relativeHeight)
    end
    f.GetRelativeWidth = function()
        return f.relativeWidth
    end
    f.GetRelativeHeight = function()
        return f.relativeHeight
    end

    f.isRelativeSizesPatched = true
end

function LootSquid:PatchEvents(f)
    if f.lsEvents == nil then
        f.lsEvents = {}
    end

    -- TODO hook events etc and wrap to make something that can be removed during runtime etc.
end

function LootSquid:CreateFrame(frameType, name, parent, template)
    local f
    if name ~= nil then
        f = CreateFrame(frameType, "LootSquid_frames_" .. name, parent, template)
    else
        f = CreateFrame(frameType, "LootSquid_frames_auto_"
                .. tostring(LootSquid.frames.autoIdCnt), parent, template)

        LootSquid.frames.autoIdCnt = LootSquid.frames.autoIdCnt + 1
    end

    LootSquid:PatchRelativeSizing(f)
    LootSquid:PatchRelativeAnchors(f)
    LootSquid:PatchUpdateLayout(f)
    LootSquid:PatchEvents(f)

    return f
end

function LootSquid:CreateBackdropSolid(edgeSize, insets)

    local edgeFile = "Interface\\ChatFrame\\ChatFrameBackground"

    -- backdrops can't handle edgesize of zero with edgeFile defined
    if edgeSize < 1 then
        edgeFile = nil
    end

    local backdrop = {
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
        edgeFile = edgeFile,
        tile = true,
        -- size (width or height) of the square repeating background tiles (in pixels)
        tileSize = 32,
        -- thickness of edge segments and square size of edge corners (in pixels)
        edgeSize = edgeSize,
        -- distance from the edges of the frame to those of the background texture (in pixels)
        insets = insets
    }

    return backdrop
end

function LootSquid:CreateButtonStandard(parent, width, height, text)
    local b = CreateFrame("Button", nil, parent, "UIPanelButtonTemplate")
    b:SetText(text)
    b:SetWidth(width)
    b:SetHeight(height)
    b:ClearAllPoints()

    LootSquid:PatchRelativeAnchors(b)
    LootSquid:PatchRelativeSizing(b)
    LootSquid:PatchUpdateLayout(b)

    return b
end

function LootSquid:CreateFrameSolid(parent, width, height, backgroundColor, edgeColor, edgeSize, isMoveAble)
    local f = LootSquid:CreateFrame("Frame", nil, parent)
    --local f = CreateFrame("Frame", nil, parent)
    if width then
        if width > 1.0 then
            f:SetWidth(width)
        else
            --print("rel1: "..tostring(width))
            f:SetRelativeWidth(width)
            --print("rel2")
        end
    end
    if height then
        if height > 1.0 then
            f:SetHeight(height)
        else
            f:SetRelativeHeight(height)
        end
    end
    f:SetBackdrop(self:CreateBackdropSolid(edgeSize))
    f:SetBackdropColor(backgroundColor.r, backgroundColor.g, backgroundColor.b, backgroundColor.a)
    f:SetBackdropBorderColor(edgeColor.r, edgeColor.g, edgeColor.b, edgeColor.a)

    if isMoveAble then

        f:SetMovable(true)
        f:EnableMouse(true)

        f:SetScript("OnMouseDown",
            function(selfFrame, button)
                if button == "LeftButton" and not selfFrame.isMoving then
                    selfFrame:StartMoving()
                    selfFrame.isMoving = true
                end
            end)

        f:SetScript("OnMouseUp",
            function(selfFrame, button)
                if button == "LeftButton" and selfFrame.isMoving then
                    selfFrame:StopMovingOrSizing()
                    selfFrame.isMoving = false
                end
            end)

        f:SetScript("OnHide",
            function(selfFrame, button)
                if selfFrame.isMoving then
                    selfFrame:StopMovingOrSizing()
                    selfFrame.isMoving = false
                end
            end)
    end

    return f
end

function LootSquid:CreateLabelStandard(parent, text, textHeight, font, isWrapDisabled)

    if font == nil then
        font = "GameFontNormal"
    end

    local label = parent:CreateFontString(nil, "OVERLAY", font)
    label:SetText(text)
    if textHeight ~= nil then
        label:SetTextHeight(textHeight)
    end

    if isWrapDisabled == true then
        label:SetWordWrap(false)
    else
        label:SetWordWrap(true)
    end

    LootSquid:PatchRelativeAnchors(label)
    LootSquid:PatchRelativeSizing(label)
    LootSquid:PatchUpdateLayout(label)


    return label
end

function LootSquid:CreateScrollFrame(frame, itemHeight, itemCnt, padding, funcItem)

    local result = {}

    local frameScroll = CreateFrame("ScrollFrame", nil, frame)
    frameScroll:SetPoint("TOPLEFT", 10, -20)
    frameScroll:SetPoint("BOTTOMRIGHT", -10, 10)

    result.frameScroll = frameScroll

    local texture = frameScroll:CreateTexture()
    texture:SetAllPoints()
    texture:SetTexture(0, 0, 0, 1)

    local scrollbar = CreateFrame("Slider", nil, frameScroll, "UIPanelScrollBarTemplate")
    scrollbar:SetPoint("TOPLEFT", frame, "TOPRIGHT", 4, -16)
    scrollbar:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 4, 16)

    --local scrollBarMax = math.max((itemHeight+padding) * itemCnt - frameScroll:GetHeight() - padding, 1)
    local scrollBarMax = math.max((itemHeight + padding) * itemCnt - padding, 1)

    scrollbar:SetMinMaxValues(0, scrollBarMax + 1)
    scrollbar:SetValueStep(itemHeight + padding)
    scrollbar.scrollStep = itemHeight + padding
    scrollbar:SetValue(0)
    scrollbar:SetWidth(16)
    scrollbar:SetScript("OnValueChanged",
        function(self, value)
            self:GetParent():SetVerticalScroll(value)
        end)
    scrollbar:EnableMouseWheel(true)
    result.scrollbar = scrollbar

    frameScroll:SetScript("OnMouseWheel",
        function(self, delta)

            local current = self:GetVerticalScroll()
            local min, max = scrollbar:GetMinMaxValues()

            local offsetCnt = math.floor(current / (itemHeight + padding))
            offsetCnt = offsetCnt - delta

            local offset = offsetCnt * (itemHeight + padding)

            if offset >= min and offset <= max then
                scrollbar:SetValue(offset)
            end
        end)

    local scrollbg = scrollbar:CreateTexture(nil, "BACKGROUND")
    scrollbg:SetAllPoints(scrollbar)
    scrollbg:SetTexture(0, 0, 0, 0.4)

    local lootFrameScrollChild = CreateFrame("Frame", nil, frameScroll)
    lootFrameScrollChild:SetWidth(frameScroll:GetWidth())
    lootFrameScrollChild:SetHeight(itemHeight * itemCnt)

    frameScroll:SetScrollChild(lootFrameScrollChild)


    for i = 0, itemCnt - 1 do
        local lootItemHolder = self:CreateFrameSolid(lootFrameScrollChild,
            lootFrameScrollChild:GetWidth(),
            itemHeight,
            { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            0,
            false)
        lootItemHolder:SetPoint("TOP", 0, -(padding + itemHeight) * i)

        funcItem(lootItemHolder, i)
    end


    result.HideIndexAndAfter = function(self, frameScroll, index)

        for i = index, LootSquid:tablelength(frameScroll.items) - 1 do
            frameScroll.items[i].frame:Hide()
        end

        local scrollBarMax = math.max((itemHeight + padding) * index - padding, 1)
        frameScroll.scrollbar:SetMinMaxValues(0, scrollBarMax + 1)
    end


    local frameMessage = self:CreateLabelStandard(frameScroll, "No more items", nil, "GameFontWhiteTiny")
    frameMessage:SetPoint("TOP", 0, -20)

    result.frameMessage = frameMessage

    return result
end


