function LootSquid:TestRunAll()

    --[[ add tests here ]]--
    local tests = {
        ["TestOnChatMessage"] = self.TestOnChatMessage
    }



    --[[ setup ]]--
    LootSquid.debug.alwaysParseLoot = true
    LootSquid.debug.alwaysParseChat = true
    LootSquid.debug.bypassSafeCalls = true
    LootSquid.debug.disableWhispers = true
    LootSquid.debug.logging = true
    LootSquid.session.currentLootSlot = 1


    LootSquid:ClearNeedAndGreedRequests()

    _G.LootSquid.session.loot[1] = {
        itemName = "Broken Fang",
        itemQuantity = nil,
        itemQuality = nil,
        itemLocked = nil,
        itemLootType = nil,
        itemLink = "|cff9d9d9d|Hitem:7073:0:0:0:0:0:0:0:80:0:0|h[Broken Fang]|h|r",
        candidates = {}
    }

    local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice =
    GetItemInfo("|cff9d9d9d|Hitem:113612:0:0:0:0:0:0:0:100:0:3:1:560|h[Scales of Doom]|h|r")


    LootSquid.session.id = 1
    LootSquid.session.currentLootSlot = 1
    LootSquid.session.currentLootOffered = {
        name = name,
        link = link,
        quality = quality,
        iLevel = iLevel,
        class = class,
        subclass = subclass,
        maxStack = maxStack,
        equipSlot = equipSlot,
        texture = texture
    }


    print(equipSlot)

    --[[ run all tests ]]--
    print("opening loot gui")
    LootSquid:OpenLootGui()

    local totalTestsSuccess = 0
    local totalTestsFail = 0

    for k,v in pairs(tests) do
        local success, error = pcall(v)
        local functionName = tostring(k)
        if not success then
            totalTestsFail = totalTestsFail + 1
            LootSquid:LogError(
                "Test failed. Name: "..functionName
                        ..", Error: "..tostring(error)
            )
        else
            totalTestsSuccess = totalTestsSuccess + 1
            LootSquid:LogInfo(
                string.format("|cFF00FF00%s:|r ", "Test succeeded") .. "Name: "..functionName
            )
        end
    end

    local totalTests = totalTestsFail + totalTestsSuccess

    if totalTestsFail > 0 then
        LootSquid:LogError(
            string.format("|cFFFF0000%s:|r ", "Test summary. ")
                    ..tostring(totalTestsFail)
                    .." out of "
                    ..tostring(totalTests)
                    .." failed."
        )
    else
        LootSquid:LogInfo(
            string.format("|cFF00FF00%s:|r ", "Test summary. ")
                    ..tostring(totalTestsSuccess)
                    .." out of "
                    ..tostring(totalTests)
                    .." succeeded."
        )
    end

    --[[ clean up ]]--
    LootSquid.debug.alwaysParseLoot = false
    LootSquid.debug.alwaysParseChat = false
    LootSquid.debug.bypassSafeCalls = false
    LootSquid.debug.disableWhispers = false
    LootSquid.debug.logging = false
    LootSquid.session.currentLootSlot = 0
end

function LootSquid:TestOnChatMessage()

    local messages = {
        -- player
        {
            sender = Ambiguate(UnitName("player"), "none"),
            message = "!need |cff9d9d9d|Hitem:7073:0:0:0:0:0:0:0:80:0:0|h[Broken Fang]|h|r GIMME BIS!"
        },
        -- need with spaces
        {
            sender = "Goolt",
            message = "!need         |cff9d9d9d|Hitem:113595:0:0:0:0:0:0:0:100:0:3:0|h[Treads of Sand and Blood]|h|r"
        },

        -- need with no space
        {
            sender = "Hanslock",
            message = "!need|cff9d9d9d|Hitem:113592:0:0:0:0:0:0:0:100:0:3:0|h[Bileslinger's Censer]|h|r"
        },

        -- need with unneeded characters
        {
            sender = "Fnitzer",
            message = "!need dgd |cff9d9d9d|Hitem:113612:0:0:0:0:0:0:0:100:0:3:1:560|h[Scales of Doom]|h|r"
        },

        -- need from other realm
        {
            sender = "Zulgarip-Dentarg",
            message = "!need |cff9d9d9d|Hitem:113641:0:0:0:0:0:0:0:100:0:3:0|h[Living Mountain Shoulderguards]|h|r"
        },

        -- first need
        {
            sender = "Zayne",
            message = "!need |cff9d9d9d|Hitem:113838:0:0:0:0:0:0:0:100:0:3:0|h[Gar'tash, Hammer of the Breakers]|h|r"
        },

        -- then change to greed
        {
            sender = "Zayne",
            message = "!greed |cff9d9d9d|Hitem:113838:0:0:0:0:0:0:0:100:0:3:0|h[Gar'tash, Hammer of the Breakers]|h|r"
        },

        -- and an actual greed
        {
            sender = "Theowan",
            message = "!greed |cff9d9d9d|Hitem:113857:0:0:0:0:0:0:0:100:0:3:0|h[Staff of the Grand Imperator]|h|r"
        },

        -- greed with no link (which should replace the linked item with no item at all)
        {
            sender = "Tryllefisa",
            message = "!greed GIEF NAUW"
        },
    }

    for k,v in pairs(messages) do
        LootSquid:OnChatMessage(v.message, v.sender)
    end

    -- all events will be processed on next update, hence there is no difference in time. we fix by altering update time
    LootSquid.lastChatRequestEventTime = GetTime() + 1.0
end
