function LootSquid:SafeCall(func, ...)
    print("safecall.." .. (...))

    if func == nil then
        LootSquid:LogError("nil is not accepted as function argeument.")
        return
    end

    if LootSquid.debug.bypassSafeCalls then
        return true, func(...)
    else
        return pcall(func, ...)
    end
end

function LootSquid:LogDebug(message, verbosity)
    if LootSquid.debug.logging then
        DEFAULT_CHAT_FRAME:AddMessage(format("|cFFFFAA00%s:|r %s", "LootSquid [Debug]", tostring(message)))
    end
end

function LootSquid:LogInfo(message)
    DEFAULT_CHAT_FRAME:AddMessage(format("|cFFFFAA00%s:|r %s", "LootSquid [Info]", tostring(message)))
end

function LootSquid:LogError(message)
    DEFAULT_CHAT_FRAME:AddMessage(format("|cFFFFAA00%s:|cFFFF0000 %s", "LootSquid [Error]", tostring(message)))
end

function LootSquid:strStartsWith(String, Start)
    return string.sub(String, 1, string.len(Start)) == Start
end

function LootSquid:strEndsWith(String, End)
    return End == '' or string.sub(String, -string.len(End)) == End
end

-- Sort a Table[] in Lua
-- http://stackoverflow.com/a/15706820
function LootSquid:spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys + 1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys
    if order then
        table.sort(keys, function(a, b)
            if a == nil and b == nil then
                return false
            end
            if a == nil then
                return false
            end
            if b == nil then
                return true
            end
            return order(t, a, b)
        end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

-- http://stackoverflow.com/a/2705804
function LootSquid:tablelength(T)
    local count = 0
    for _ in pairs(T) do count = count + 1 end
    return count
end

function LootSquid:IsPlayerAllowedToOfficerChat()
    local _, _, rank = GetGuildInfo("player");

    if not rank then
        return false
    end

    GuildControlSetRank(rank + 1)
    local guildchat_listen, guildchat_speak, officerchat_listen, officerchat_speak = GuildControlGetRankFlags()
    return officerchat_listen and officerchat_speak
end

function LootSquid:IsPlayerMasterLooter()

    if (not UnitInRaid("player")) then
        return false
    end

    local master, partyMaster, raidMaster = GetLootMethod()
    local playerName = UnitName("player")

    local raidName = "raid" .. tostring(raidMaster)

    if master and raidMaster and playerName == UnitName(raidName) then
        self:LogDebug("IsPlayerMasterLooter: " .. tostring(true))
        return true
    else
        self:LogDebug("IsPlayerMasterLooter: " .. tostring(false))
        return false
    end
end

function LootSquid:GetMode()

    --return "Council"
--return "MasterLooter"

    if LootSquid:IsPlayerMasterLooter() then
        return "MasterLooter"
    elseif LootSquid:IsPlayerAllowedToOfficerChat() then
        return "Council"
    else
        return "Player"
    end
end

function LootSquid:TimeToDateTable(time)
    return date("*t", tonumber(time))
end

function LootSquid:TimeToDateTable(time)
    return date("!*t", tonumber(time))
end

function LootSquid:TimeToDateStr(time)
    return date("%Y-%m-%d %H:%M:%S", time)
end

function LootSquid:TimeToUtcDateStr(time)
    return date("!%Y-%m-%d %H:%M:%S", time)
end

function LootSquid:GetDateDifference(a, b)
    return a.year - b.year, a.month - b.month, a.day - b.day, a.hour - b.hour, a.min - b.min, a.sec - b.sec
end

function LootSquid:GetDateDifferenceTotalDays(a, b)
    local diff = time(a) - time(b)
    return diff / 86400.0, diff
end

local LOOTSQUID_INVTYPES = {
    ["INVTYPE_HEAD"] = GetInventorySlotInfo("HeadSlot"),
    ["INVTYPE_NECK"] = GetInventorySlotInfo("NeckSlot"),
    ["INVTYPE_SHOULDER"] = GetInventorySlotInfo("ShoulderSlot"),
    ["INVTYPE_CHEST"] = GetInventorySlotInfo("ChestSlot"),
    ["INVTYPE_ROBE"] = GetInventorySlotInfo("ChestSlot"),
    ["INVTYPE_WAIST"] = GetInventorySlotInfo("WaistSlot"),
    ["INVTYPE_LEGS"] = GetInventorySlotInfo("LegsSlot"),
    ["INVTYPE_FEET"] = GetInventorySlotInfo("FeetSlot"),
    ["INVTYPE_WRIST"] = GetInventorySlotInfo("WristSlot"),
    ["INVTYPE_HAND"] = GetInventorySlotInfo("HandsSlot"),
    ["INVTYPE_FINGER"] = {GetInventorySlotInfo("Finger0Slot"), GetInventorySlotInfo("Finger1Slot")},
    ["INVTYPE_TRINKET"] = {GetInventorySlotInfo("Trinket0Slot"), GetInventorySlotInfo("Trinket1Slot")},
    ["INVTYPE_CLOAK"] = GetInventorySlotInfo("BackSlot"),
    ["INVTYPE_WEAPON"] = {GetInventorySlotInfo("MainHandSlot"), GetInventorySlotInfo("SecondaryHandSlot")},
    ["INVTYPE_SHIELD"] = GetInventorySlotInfo("SecondaryHandSlot"),
    ["INVTYPE_2HWEAPON"] = GetInventorySlotInfo("MainHandSlot"),
    ["INVTYPE_WEAPONMAINHAND"] = GetInventorySlotInfo("MainHandSlot"),
    ["INVTYPE_WEAPONOFFHAND"] = GetInventorySlotInfo("SecondaryHandSlot"),
    ["INVTYPE_HOLDABLE"] = GetInventorySlotInfo("SecondaryHandSlot"),
    --["INVTYPE_RANGED"] = GetInventorySlotInfo("RangedSlot"),
    --["INVTYPE_THROWN"] = GetInventorySlotInfo("RangedSlot"),
    --["INVTYPE_RANGEDRIGHT"] = GetInventorySlotInfo("RangedSlot"),
    --["INVTYPE_RELIC"] = GetInventorySlotInfo("RangedSlot"),
}


function LootSquid:GetInvItemsByInvType(invType)
    local invSlots = LOOTSQUID_INVTYPES[invType]

    local availableItems = {}

    if type(invSlots) == "table" then
        GetInventoryItemsForSlot(invSlots[1], availableItems)
    else
        GetInventoryItemsForSlot(invSlots, availableItems)
    end

    return availableItems
end

LootSquid.components.utils = true
